#include "main.h"

int 
main(int argc, char **argv) 
{
  command_line_process(argc, argv);

  init_signal_handlers();
  init_time();

  read_config_file();
  print_banner();
  read_ban_file();

  init_filler_packet();
  init_quake_connection();
  init_client_connections();

  net_init();

  if( gs.accepting_clients ) {
    open_listen_connection();
  }
  connect_quake_server();

  run_main_loop();
  return 0;
}

void
run_main_loop(void)
{
  int i;

  restart_timer(&gs.pollinput_timer);
  restart_timer(&gs.retry_timer);

  while( TRUE ) {
    net_wait_for_any_input(POLL_TIMEOUT);

    if( is_connected() && net_check_for_input(gs.server_socket) ) {
      process_server_packet();
    }

    if( read_timer(&gs.pollinput_timer) >= POLL_TIMEOUT ) {
      do_timeout_ops();
      restart_timer(&gs.pollinput_timer);
    }

    if( !is_in_the_game() &&
        read_timer(&gs.retry_timer) >= SECS_BETWEEN_RETRYS ) 
    {
      connect_quake_server();
      restart_timer(&gs.retry_timer);
    }

    if( gs.accepting_clients ) {
      if( net_check_for_input(gs.listen_socket) ) {
        accept_new_client();
      }

      for(i=0; i<MAX_CLIENTS; i++) {
        if( gs.client[i].valid && net_check_for_input(gs.client[i].socket) ) {
          recv_client_message(i);
        }
      }

      for(i=0; i<MAX_CLIENTS; i++) {
        if( gs.client[i].valid && !gs.client[i].password &&
            read_timer(&gs.client[i].password_timer) >= PASSWORD_TIMEOUT ) {
          send_to_client(i, "commandprint No password supplied within time limit\n");
          disconnect_client(i);
        }
      }
    }

  }
}

void 
command_line_process(int argc, char **argv)
{
    int   mode = 0, gotConfigFile = 0;
    char *prog, *tok;

    prog = (tok = strrchr(argv[0], '/')) ? &tok[1] : argv[0];

    argc--;
    argv++;

    while( argc ) {
        if( strcmp(argv[0], "-debug") == 0 ) {
            if (argc >= 2) {
                if( strcmp(argv[1], "netin") == 0 ) {
                    mode |= DEBUG_INCOMING_PACKET_DATA;
                } else if( strcmp(argv[1], "netout") == 0 ) {
                    mode |= DEBUG_OUTGOING_PACKET_DATA;
                } else if( strcmp(argv[1], "tcpin") == 0 ) {
                    mode |= DEBUG_INCOMING_TCP_DATA;
                } else if( strcmp(argv[1], "tcpout") == 0 ) {
                    mode |= DEBUG_OUTGOING_TCP_DATA;
                } else if( strcmp(argv[1], "cmdin") == 0 ) {
                    mode |= DEBUG_INCOMING_COMMANDS;
                } else if( strcmp(argv[1], "cmdout") == 0 ) {
                    mode |= DEBUG_OUTGOING_COMMANDS;
                } else {
                    fprintf(stderr, "%s: unrecognized debug mode '%s'\n", 
                      prog, argv[1]);
                    usage();
                }
            } else {
              usage();
            }
        } else if( strcmp(argv[0], "-f") == 0 ) {
            if (argc >= 2) {
                strcpy(gs.configfile, argv[1]);
                gotConfigFile = 1;
            } else {
              usage();
            }
        } else if( strcmp(argv[0], "help") == 0 || strcmp(argv[0], "-help") == 0 ) {
           usage();
        } else {
           fprintf(stderr, "%s: Unrecognized option '%s'\n", prog, argv[0]);
           usage();
        }
        argc -= 2;
        argv += 2;
    }

    if( !gotConfigFile )  {
        fprintf(stderr, "%s: A config file must be specified.\n\n", prog);
        usage();
    }

    debug_set_file(stderr);
    debug_set_mode(mode);
}

void 
shutdown_all()
{
   send_disconnect();
   disconnect_all_clients();
   destroy_stuff_commands();
   destroy_ban_lists();
   fclose(gs.logfp);
   exit(1);
}

void 
print_to_log(char *fmt, ...)
{
  va_list arglist;
  
  va_start(arglist, fmt); 
  fprintf(gs.logfp, "%s ", gs.timestr);
  vfprintf(gs.logfp, fmt, arglist);
  fflush(gs.logfp);
  va_end(arglist);
}

void 
print_to_stats_log(char *fmt, ...)
{
  va_list arglist;
  
  if( gs.statslogfp ) {
    va_start(arglist, fmt); 
    vfprintf(gs.statslogfp, fmt, arglist);
    fflush(gs.statslogfp);
    va_end(arglist);
  }
}

void 
init_time(void)
{
  time_t tp;

  time(&tp);
  strftime(gs.timestr, sizeof(gs.timestr),"%c", localtime(&tp));
}

void 
do_timeout_ops(void)
{
  time_t tp;

  time(&tp);
  strftime(gs.timestr, sizeof(gs.timestr),"%c", localtime(&tp));

  send_reliable_head();

  if( is_in_the_game() ) {
    if( ban_ok() ) {
      send_impulse(gs.ban_impulse);
    }
    send_console_command("status");
    send_to_clients("time %f\n", gs.time);
  }
}

void 
sig_catcher(int signo)
{
  print_to_log("Caught signal: %d\n", signo);
  shutdown_all();
}

void 
init_signal_handlers()
{
#ifndef WIN32
  signal(SIGINT,  sig_catcher);
  signal(SIGQUIT, sig_catcher);
  signal(SIGBUS,  sig_catcher);
  signal(SIGSEGV, sig_catcher);
  signal(SIGTERM, sig_catcher);
  signal(SIGPIPE, SIG_IGN);
#endif
}

void 
usage()
{
  fprintf(stderr, 
   "usage: qsmack -f file [-debug option -help]\n\n"

   " This is version %.2f\n\n"

   " -f file\n"
   "                   Specifies the name of the config file. See our website:\n"
   "                   http://lemur.stanford.edu/clan9/qsmack for documentation,\n"
   "                   information and updates.\n\n"

   " -debug netin|netout|tcpin|tcpout|cmdin|cmdout\n"
   "                   Multiple debug options be give with multiple -debug\n"
   "                   switches. Default is no debug.\n\n",
   VERSION
  );
  exit(1);
}

void 
print_banner()
{
  fprintf(gs.logfp,
    "*******************************************************************************\n"
    "*                              Welcome to QSmack                              *\n"
    "*                                Version %.2f                                 *\n"
    "*                  Brought to you by Clan 9 from Outer Space                  *\n"
    "*                  See http://lemur.stanford.edu/clan9/qsmack                 *\n" 
    "*                   for documention, information and updates                  *\n"
    "*******************************************************************************\n",
    VERSION
  );
}
